#!/usr/bin/python
# -*- coding: utf-8 -*-

from io import BytesIO
from shutil import rmtree
from typing import IO, Any, Generator, List, Literal
from uuid import uuid4
from time import sleep
from contextlib import contextmanager
from os import makedirs, listdir, rename, remove, environ
from os.path import exists, abspath, isabs
from selenium.webdriver import Chrome, ChromeOptions
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait, Select

MAX_WAIT_TIME = 30.0

ImageFormat = Literal["png", "jpeg", "gif", "bmp", "tiff", "svg"]


class WebyOptions:
    headless: bool = False
    stable: bool = True

    def __init__(self, headless: bool = False, stable: bool = True):
        self.headless = headless
        self.stable = stable


class WebyAPI:
    _driver: Chrome | None = None
    _wait: WebDriverWait[Chrome] | None = None
    _remove_downloads: bool = False
    _downloads_dir: str = abspath("downloads")
    _temp_dir: str = abspath("temp")
    _last_path: str | None = None

    @classmethod
    def build_options(cls):
        return WebyOptions(headless=bool(environ.get("HEADLESS", False)))

    def start(self, options: WebyOptions = WebyOptions()):
        if not exists(self._downloads_dir):
            makedirs(self._downloads_dir)

        if not exists(self._temp_dir):
            makedirs(self._temp_dir)

        chrome_options = ChromeOptions()
        if options.headless:
            chrome_options.arguments.append("--headless=new")
        if options.stable:
            chrome_options.arguments.append("--no-sandbox")
            chrome_options.arguments.append("--disable-dev-shm-usage")
            chrome_options.arguments.append("--disable-gpu")
            chrome_options.arguments.append("--disable-setuid-sandbox")
            chrome_options.arguments.append("--remote-debugging-port=9222")
        chrome_options.add_experimental_option(
            "prefs",
            {
                "download.default_directory": self._temp_dir,
                "download.prompt_for_download": False,
                "download.directory_upgrade": True,
                "safebrowsing.enabled": False,
                "safebrowsing.disable_download_protection": True,
            },
        )

        self._driver = Chrome(options=chrome_options)
        if options.stable:
            self._driver.delete_all_cookies()

        self._wait = WebDriverWait(self._driver, MAX_WAIT_TIME)

    def stop(self):
        if self.driver is not None:
            self.driver.quit()
        if self._remove_downloads and exists(self._downloads_dir):
            rmtree(self._downloads_dir)

    def get_element(self, by: str, value: str) -> WebElement:
        return self.wait.until(
            expected_conditions.presence_of_element_located((by, value))
        )

    def get_elements(self, by: str, value: str) -> List[WebElement]:
        return self.wait.until(
            expected_conditions.presence_of_all_elements_located((by, value))
        )

    def select_item(
        self,
        element: WebElement,
        text: str,
        force: bool = False,
        timeout: float | None = None,
    ) -> Select:
        select = Select(element)
        selected_text = select.first_selected_option.text
        if not force and selected_text == text:
            return select
        select.select_by_visible_text(text)
        if timeout is not None:
            sleep(timeout)
        return select

    def select_item_index(
        self,
        element: WebElement,
        index: int,
        force: bool = False,
        timeout: float | None = None,
    ) -> Select:
        select = Select(element)
        selected_index = select.options.index(select.first_selected_option)
        if not force and selected_index == index:
            return select
        select.select_by_index(index)
        if timeout is not None:
            sleep(timeout)
        return select

    def wait_download(
        self,
        file_path: str | None = None,
        suffix: str = ".crdownload",
        timeout: float = MAX_WAIT_TIME,
        step_time: float = 1.0,
        move_file: bool = True,
        overwrite: bool = True,
    ) -> str | None:
        seconds = 0.0
        dst: str | None = None

        while seconds < timeout:
            sleep(step_time)
            seconds += step_time

            # in case the file is not yet in temp download, then
            # the download is yet to be started
            if not listdir(self._temp_dir):
                continue

            # in case there's a file with the suffix, then the download
            # is ongoing (need to wait a little bit more)
            if any([filename.endswith(suffix) for filename in listdir(self._temp_dir)]):
                continue

            # if the file should be moved, then move it from
            # the temp download folder to the downloads folder
            if move_file:
                filename = listdir(self._temp_dir)[0]

                if filename in listdir(self._downloads_dir):
                    if overwrite:
                        remove(f"{self._downloads_dir}/{filename}")
                    else:
                        raise Exception(
                            f"File {filename} already exists in downloads folder"
                        )

                src = f"{self._temp_dir}/{filename}"
                dst = f"{self._downloads_dir}/{filename}"

                if file_path is not None:
                    dst = (
                        file_path
                        if isabs(file_path)
                        else f"{self._downloads_dir}/{file_path}"
                    )

                src = abspath(src)
                dst = abspath(dst)

                rename(src, dst)

            self._last_path = dst

            # if we reach this point, then the download is completed
            # we can return the destination file path
            return dst

        raise Exception(f"Download not completed after {timeout} seconds")

    def screenshot(self) -> bytes:
        return self.driver.get_screenshot_as_png()

    def screenshot_file(
        self, name: str | None = None, format: ImageFormat = "png"
    ) -> str:
        if name is None:
            name = f"{uuid4()}"

        filename = f"{name}.{format}"

        if not self.driver.get_screenshot_as_file(filename):
            raise Exception(f"Failed to save screenshot to {filename}")

        return filename

    @contextmanager
    def driver_ctx(
        self, options: WebyOptions | None = None, stop=True
    ) -> Generator[Chrome, Any, Any]:
        if options is None:
            options = self.build_options()
        self.start(options=options)
        try:
            yield self.driver
        finally:
            if stop:
                self.stop()

    @contextmanager
    def download_ctx(
        self, wait_download=True, file_path: str | None = None
    ) -> Generator[str, Any, Any]:
        self._cleanup_temp()
        try:
            yield self._downloads_dir
        finally:
            if wait_download:
                self.wait_download(file_path=file_path)

    @property
    def driver(self) -> Chrome:
        if self._driver is None:
            raise Exception("Driver is not started")
        return self._driver

    @property
    def wait(self) -> WebDriverWait[Chrome]:
        if self._wait is None:
            raise Exception("Wait is not started")
        return self._wait

    def _cleanup_temp(self):
        for filename in listdir(self._temp_dir):
            remove(f"{self._temp_dir}/{filename}")

    def _last_download(self) -> IO:
        return open(self._last_download_path, "rb")

    def _last_download_buffer(self) -> IO:
        file = self._last_download()
        try:
            content = file.read()
        finally:
            file.close()
        buffer = BytesIO(content)
        buffer.write(content)
        buffer.seek(0)
        return buffer

    @property
    def _last_download_path(self) -> str:
        if self._last_path is None:
            raise Exception("No file downloaded")
        if not exists(self._last_path):
            raise Exception(f"File {self._last_download} does not exist")
        return self._last_path



#!/usr/bin/python
# -*- coding: utf-8 -*-

from enum import Enum


class FileType(Enum):
    PDF = "application/pdf"
    CSV = "text/csv"
    TXT = "text/plain"

    @property
    def extension(self) -> str:
        return self.name.lower()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from enum import Enum
from time import sleep
from os import environ
from os.path import basename
from datetime import datetime
from typing import IO, Literal, Sequence, cast
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import Select

from .base import WebyAPI
from .common import FileType

BpiSections = Literal[
    "Consultas",
    "Operações",
    "Ficheiros",
    "Internacional",
    "Financiamento",
    "Factoring e Confirming",
    "Cartões",
    "TPA",
    "Investimento",
    "Autorizações",
]

BpiSideSections = Literal[
    "Posição Integrada",
    "Posição Integrada Global",
    "Saldos",
    "Lista Saldos",
    "Movimentos",
    "Avisos/Faturas/Notas Crédito e Débito",
    "Avaliação da Adequação de Prod. Investimento",
    "Extrato Conta",
    "Extrato Cartões",
]

ReportSections = Literal["Extrato Conta", "Extrato Investimento"]

SelectDateRange = Literal[
    "Últimos 3 dias",
    "Última Semana",
    "Último Mês",
    "Últimos 3 Meses",
    "Último Ano",
    "Intervalo de Datas",
]

DocumentType = Literal[
    "Todos",
    "Avisos",
    "Factura-Recibo",
    "Facturas",
    "Notas de Crédito",
    "Notas de Débito",
]


class BpiAPI(WebyAPI):
    username: str | None = None
    password: str | None = None

    def __init__(self, username: str | None = None, password: str | None = None):
        super().__init__()

        self.username = username
        self.password = password
        self.build_login()

    def build_login(self):
        if self.username is None:
            self.username = environ.get("BPI_USERNAME", None)
        if self.password is None:
            self.password = environ.get("BPI_PASSWORD", None)

        if self.username is None:
            raise Exception("BPI_USERNAME must be set")
        if self.password is None:
            raise Exception("BPI_PASSWORD must be set")

    def get_balance(self) -> str:
        with self.driver_ctx():
            self.login()
            self.select_section("Consultas")
            self.select_side_menu("Movimentos")
            return self.text_balance()

    def download_invoice(
        self,
        date_range: SelectDateRange = "Último Ano",
        document_type: DocumentType = "Facturas",
        invoice_indexes: Sequence[int] = (0,),
    ) -> Sequence["BpiDocument"]:
        docs: list[BpiDocument] = []
        with self.driver_ctx():
            self.login()
            self.select_section("Consultas")
            self.select_side_menu("Avisos/Faturas/Notas Crédito e Débito")
            self.select_filters(date_range=date_range, document_type=document_type)
            for invoice_index in invoice_indexes:
                self.click_extract(row_index=invoice_index)
                docs.append(
                    BpiDocument(
                        BpiDocumentType.INVOICE,
                        basename(self._last_download_path),
                        self._last_download_buffer(),
                        file_type=FileType.PDF,
                        account=self.username,
                        date=datetime.strptime(
                            basename(self._last_download_path)[:10], "%Y-%m-%d"
                        ),
                    )
                )
        return docs

    def download_report(
        self,
        section: ReportSections = "Extrato Conta",
        report_indexes: Sequence[int] = (0,),
        account_index: int | None = None,
    ) -> Sequence["BpiDocument"]:
        docs: list[BpiDocument] = []
        with self.driver_ctx():
            self.login()
            self.select_section("Consultas")
            self.select_side_menu(cast(BpiSideSections, section))
            if account_index is not None:
                self.select_account(account_index)
            for report_index in report_indexes:
                self.click_extract(row_index=report_index)
                docs.append(
                    BpiDocument(
                        BpiDocumentType.from_section(section),
                        basename(self._last_download_path),
                        self._last_download_buffer(),
                        file_type=FileType.PDF,
                        account=self.username,
                        date=datetime.strptime(
                            basename(self._last_download_path)[-14:-4], "%Y-%m-%d"
                        ),
                    )
                )
        return docs

    def download_account_report(
        self, report_indexes: Sequence[int] = (0,), account_index: int = 0
    ) -> Sequence["BpiDocument"]:
        return self.download_report(
            section="Extrato Conta",
            report_indexes=report_indexes,
            account_index=account_index,
        )

    def download_investing_report(
        self, report_indexes: Sequence[int] = (0,)
    ) -> Sequence["BpiDocument"]:
        return self.download_report(
            section="Extrato Investimento", report_indexes=report_indexes
        )

    def download_card_report(
        self, card_index=0, report_indexes: Sequence[int] = (0,)
    ) -> Sequence["BpiDocument"]:
        docs: list[BpiDocument] = []
        with self.driver_ctx():
            self.login()
            self.select_section("Consultas")
            self.select_side_menu("Extrato Cartões")
            for report_index in report_indexes:
                self.click_card_account(row_index=card_index)
                self.click_extract(row_index=report_index)
                docs.append(
                    BpiDocument(
                        BpiDocumentType.CARD_EXTRACT,
                        basename(self._last_download_path),
                        self._last_download_buffer(),
                        file_type=FileType.PDF,
                        account=self.username,
                        date=datetime.strptime(
                            basename(self._last_download_path)[-14:-4], "%Y-%m-%d"
                        ),
                    )
                )
        return docs

    def login(self, username: str | None = None, password: str | None = None):
        self.driver.get("https://bpinetempresas.bancobpi.pt/SIGNON/signon.asp")

        close = self.driver.find_element(By.ID, "fechar")
        close.click()

        username_e = self.get_element(By.XPATH, "//*[@label='Nome Acesso']")
        password_e = self.get_element(By.XPATH, "//*[@label='Código Secreto']")
        username_e.send_keys(username or self.username or "")
        password_e.send_keys(password or self.password or "")
        password_e.send_keys(Keys.RETURN)

    def text_balance(self) -> str:
        balance = self.get_element(
            By.XPATH, "//div[text()='Saldo Disponível:']/following-sibling::div/span"
        )
        return balance.text

    def select_section(self, section: BpiSections):
        section_e = self.get_element(By.XPATH, f"//a[contains(text(), '{section}')]")
        section_e.click()

    def select_side_menu(self, side_section: BpiSideSections):
        side_section_e = self.get_element(
            By.XPATH, f"//div[contains(text(), '{side_section}')]"
        )
        side_section_e.click()

    def select_filters(self, date_range: SelectDateRange, document_type: DocumentType):
        self.select_item(self.get_elements(By.XPATH, "//select")[2], date_range)
        self.select_item(self.get_elements(By.XPATH, "//select")[3], document_type)

        filter = self.get_element(By.XPATH, "//*[@value='Filtrar']")
        filter.click()

    def select_account(self, index: int = 0, timeout=5.0):
        account_element = self.get_element(
            By.XPATH, "//div[text()='Conta']/following-sibling::div/*/select"
        )
        self.select_item_index(account_element, index, timeout=timeout)

    def click_extract(self, row_index=0, wait_download: bool = True):
        open_extract = self.get_element(
            By.XPATH,
            f"//table[contains(@class, 'TableRecords')]//tr[{row_index + 1}]//a[contains(text(), 'Abrir')]",
        )
        with self.download_ctx(wait_download=wait_download):
            open_extract.click()

    def click_card_account(self, row_index=0, wait_download: bool = True):
        card_account = self.get_element(
            By.XPATH, f"//a[contains(@class, 'Text_NoWrap')][{row_index + 1}]"
        )
        card_account.click()


class BpiDocumentType(Enum):
    ACCOUNT_EXTRACT = 1
    CARD_EXTRACT = 2
    INVESTMENT_EXTRACT = 3
    INVOICE = 4
    UNKNOWN = 100

    @classmethod
    def from_section(cls, section: str) -> "BpiDocumentType":
        if section == "Extrato Conta":
            return BpiDocumentType.ACCOUNT_EXTRACT
        if section == "Extrato Cartões":
            return BpiDocumentType.CARD_EXTRACT
        if section == "Extrato Investimento":
            return BpiDocumentType.INVESTMENT_EXTRACT
        if section == "Avisos/Faturas/Notas Crédito e Débito":
            return BpiDocumentType.INVOICE
        if section == "Faturas":
            return BpiDocumentType.INVOICE
        return BpiDocumentType.UNKNOWN

    def __repr__(self):
        return self.name.lower().replace("_", " ").title()


class BpiDocument:
    type: BpiDocumentType
    name: str
    buffer: IO[bytes]
    file_type: FileType
    account: str | None
    date: datetime | None

    def __init__(
        self,
        type: BpiDocumentType,
        name: str,
        buffer: IO[bytes],
        file_type: FileType,
        account: str | None = None,
        date: datetime | None = None,
    ):
        self.type = type
        self.name = name
        self.buffer = buffer
        self.file_type = file_type
        self.account = account
        self.date = date

    def __repr__(self):
        return f"BpiDocument(type={self.type}, name={self.name}, account={self.account} date={self.date})"

    @property
    def year(self) -> int:
        if self.date is None:
            raise Exception("Date is not set")
        return self.date.year

    @property
    def month(self) -> int:
        if self.date is None:
            raise Exception("Date is not set")
        return self.date.month

    @property
    def day(self) -> int:
        if self.date is None:
            raise Exception("Date is not set")
        return self.date.day

    @property
    def month_filename(self) -> str:
        if self.date is None:
            raise Exception("Date is not set")
        return f"{self.date.strftime('%m')}.{self.file_type.extension}"



#!/usr/bin/python
# -*- coding: utf-8 -*-

from os import environ
from typing import Tuple

from .base import WebyAPI


class BigAPI(WebyAPI):
    username: str | None = None
    password: str | None = None
    nif: str | None = None

    def build_login(self):
        if self.username is None:
            self.username = environ.get("BIG_USERNAME", None)
        if self.password is None:
            self.password = environ.get("BIG_PASSWORD", None)
        if self.nif is None:
            self.nif = environ.get("BIG_NIF", None)

        if self.username is None:
            raise Exception("BIG_USERNAME must be set")
        if self.password is None:
            raise Exception("BIG_PASSWORD must be set")
        if self.nif is None:
            raise Exception("BIG_NIF must be set")
